/// <reference types="pouchdb-core" />
/// <reference types="node" />
/// <reference types="pouchdb-replication" />
/// <reference types="pouchdb-adapter-http" />
/// <reference types="pouchdb-adapter-memory" />
/// <reference types="pouchdb-adapter-leveldb" />
/// <reference types="pouchdb-adapter-cordova-sqlite" />
/// <reference types="pouchdb-adapter-fruitdown" />
/// <reference types="pouchdb-adapter-idb" />
/// <reference types="pouchdb-adapter-localstorage" />
/// <reference types="pouchdb-adapter-websql" />
/// <reference types="pouchdb-find" />
/// <reference types="pouchdb-mapreduce" />
/// <reference types="debug" />
/// <reference types="pouchdb-node" />
import { UrlObject } from 'url';
export declare type MaybeSavedPouchDoc = {
    _id?: PouchDB.Core.DocumentId;
    _rev?: PouchDB.Core.RevisionId;
};
export declare type SavedPouchDoc = {
    _id: PouchDB.Core.DocumentId;
    _rev: PouchDB.Core.RevisionId;
};
/**
 * @private
 */
declare type FirstArgument<T> = T extends (arg1: infer U, ...args: any[]) => any ? U : any;
export declare const couchUrlify: (url: string) => string;
export declare const POUCHY_API_DOCS_URI = "https://cdaringe.github.io/pouchy";
export declare type PouchyOptions = {
    conn?: UrlObject;
    couchdbSafe?: boolean;
    name?: string;
    path?: string;
    pouchConfig?: PouchDB.Configuration.DatabaseConfiguration;
    /**
     * in object form you can try `{ out/in/sync: ... }` where ... refers to the
     * [official PouchDB replication options](http://pouchdb.com/api.html#replication). in string form, simply provide
     * 'out/in/sync'. please note that the string shorthand applies default
     * heartbeat/retry options.
     */
    replicate?: string | {
        out?: PouchDB.Replication.ReplicateOptions;
        in?: PouchDB.Replication.ReplicateOptions;
        sync?: PouchDB.Replication.ReplicateOptions;
    };
    replicateLive?: boolean;
    url?: string;
};
export declare class Pouchy<Content = {}> {
    static PouchDB: PouchDB.Static;
    static plugin: (plugin: "This should be passed to PouchDB.plugin()") => PouchDB.Static;
    static defaults: (options: PouchDB.Configuration.DatabaseConfiguration) => new <Content_1 extends {} = {}>(name?: string | undefined, options?: PouchDB.Configuration.LocalDatabaseConfiguration | PouchDB.Configuration.RemoteDatabaseConfiguration | undefined) => PouchDB.Database<Content_1>;
    static debug: import("debug").Debug;
    syncEmitter: PouchDB.Replication.Replication<Content> | PouchDB.Replication.Sync<Content> | null;
    db: PouchDB.Database<Content>;
    hasLocalDb: boolean;
    isEnforcingCouchDbSafe: boolean;
    url: string | null;
    path: string | null;
    private _replicationOpts;
    constructor(opts: PouchyOptions);
    /**
     * @private
     */
    _validatePouchyOpts(opts: PouchyOptions): void;
    /**
     * @private
     */
    _handleReplication(opts: PouchyOptions['replicate']): void;
    /**
     * @private
     */
    _getUrlFromOpts(opts: PouchyOptions): string | null;
    /**
     * @private
     */
    _setDbNameFromUri(uri: string): string;
    /**
     * @private
     */
    _setDbNameFromOpts(opts: PouchyOptions): string;
    /**
     * @private
     */
    _validateDbName(): void;
    /**
     * add a document to the db.
     * @see save
     * @example
     * // with _id
     * const doc = await p.add({ _id: 'my-sauce', bbq: 'sauce' })
     * console.log(doc._id, doc._rev, doc.bbq); // 'my-sauce', '1-a76...46c', 'sauce'
     *
     * // no _id
     * const doc = await p.add({ peanut: 'butter' })
     * console.log(doc._id, doc._rev, doc.peanut); // '66188...00BF885E', '1-0d74...7ac', 'butter'
     */
    add(doc: Content & MaybeSavedPouchDoc): Promise<Content & SavedPouchDoc>;
    /**
     * get all documents from db
     * @example
     * const docs = await p.all()
     * console.log(`total # of docs: ${docs.length}!`))
     *
     * @example
     * const docs = await p.all({ includeDesignDocs: true })
     */
    all(allOpts?: FirstArgument<PouchDB.Database<Content>['allDocs']>): Promise<(Content & SavedPouchDoc)[]>;
    /**
     * The native bulkGet PouchDB API is not very user friendly.
     * In fact, it's down right wacky!
     * This method patches PouchDB's `bulkGet` and assumes that _all_ of your
     * requested docs exist.  If they do not, it will error via the usual error
     * control flows.
     *
     * @example
     * // A good example of what you can expect is actually right out of the tests!
     * let dummyDocs = [
     *   { _id: 'a', data: 'a' },
     *   { _id: 'b', data: 'b' }
     * ]
     * Promise.resolve()
     * .then(() => p.save(dummyDocs[0])) // add our first doc to the db
     * .then((doc) => (dummyDocs[0] = doc)) // update our closure doc it knows the _rev
     * .then(() => p.save(dummyDocs[1]))
     * .then((doc) => (dummyDocs[1] = doc))
     * .then(() => {
     *   // prepare getMany query (set of { _id, _rev}'s are required)
     *   const toFetch = dummyDocs.map(dummy => ({
     *     _id: dummy._id,
     *     _rev: dummy._rev
     *     // or you can provide .id, .rev
     *   }))
     *   p.getMany(toFetch)
     *     .then((docs) => {
     *       t.deepEqual(docs, dummyDocs, 'getMany returns sane results')
     *       t.end()
     *     })
     * })
     * @param {object|array} opts array of {_id, _rev}s, or { docs: [ ... } } where
     *                            ... is an array of {_id, _rev}s
     * @param {function} [cb]
     */
    getMany(docMetas: {
        _id: string;
        _rev?: string | undefined;
    }[]): Promise<(Content & SavedPouchDoc)[]>;
    /**
     * easy way to create a db index.
     * @see createIndicies
     * @example
     * await p.upsertIndex('myIndex')
     */
    upsertIndex(indexName: string): Promise<void | PouchDB.Find.CreateIndexResponse<Content>>;
    /**
     * allow single or bulk creation of indicies. also, doesn't flip out if you've
     * already set an index.
     * @example
     * const indicies = await p.createIndicies('test')
     * console.dir(indicies)
     * // ==>
     * [{
     *     id: "_design/idx-28933dfe7bc072c94e2646126133dc0d"
     *     name: "idx-28933dfe7bc072c94e2646126133dc0d"
     *     result: "created"
     * }]
     */
    createIndicies(indicies: string | string[]): Promise<void | PouchDB.Find.CreateIndexResponse<Content>>;
    /**
     * @see deleteAll
     * @returns {Promise}
     */
    clear(): Promise<PouchDB.Core.Response[]>;
    /**
     * delete a document.
     * @example
     * // same as pouch.remove
     * const deleted = await p.delete(doc)
     * console.dir(deleted)
     * // ==>
     * {
     *     id: "test-doc-1"
     *     ok: true
     *     rev: "2-5cf6a4725ed4b9398d609fc8d7af2553"
     * }
     */
    delete(doc: PouchDB.Core.RemoveDocument, opts?: PouchDB.Core.Options): Promise<PouchDB.Core.Response>;
    /**
     * clears the db of documents. under the hood, `_deleted` flags are added to docs
     */
    deleteAll(): Promise<PouchDB.Core.Response[]>;
    /**
     * Destroys the database.  Proxies to pouchdb.destroy after completing
     * some internal cleanup first
     */
    destroy(): Promise<void>;
    /**
     * Similar to standard pouchdb.find, but returns simple set of results
     */
    findMany(opts: FirstArgument<PouchDB.Find.FindRequest<Content>>): Promise<(Content & SavedPouchDoc)[]>;
    /**
     * update a document, and get your sensibly updated doc in return.
     *
     * @example
     * const doc = await p.update({ _id: 'my-doc', _rev: '1-abc123' })
     * console.log(doc)
     * // ==>
     * {
     *    _id: 'my-doc',
     *    _rev: '2-abc234'
     * }
     * @param {object} doc
     * @param {function} [cb]
     * @returns {Promise}
     */
    update(doc: FirstArgument<PouchDB.Database<Content>['put']>): Promise<Content & SavedPouchDoc>;
    /**
     * Adds or updates a document.  If `_id` is set, a `put` is performed (basic add operation). If no `_id` present, a `post` is performed, in which the doc is added, and large-random-string is assigned as `_id`.
     * @example
     * const doc = await p.save({ beep: 'bop' })
     * console.log(doc)
     * // ==>
     * {
     *   _id: 'AFEALJW-234LKJASDF-2A;LKFJDA',
     *   _rev: '1-asdblkue242kjsa0f',
     *   beep: 'bop'
     * }
     */
    save(doc: Content & MaybeSavedPouchDoc): Promise<Content & SavedPouchDoc>;
    /**
     * START POUCHDB IMPLEMENTATIONS FOR MIXIN SUPPORT
     * START POUCHDB IMPLEMENTATIONS FOR MIXIN SUPPORT
     * START POUCHDB IMPLEMENTATIONS FOR MIXIN SUPPORT
     */
    /**
     * database name
     */
    name: string;
    /** Fetch all documents matching the given options. */
    allDocs<Model>(options?: PouchDB.Core.AllDocsWithKeyOptions | PouchDB.Core.AllDocsWithKeysOptions | PouchDB.Core.AllDocsWithinRangeOptions | PouchDB.Core.AllDocsOptions): Promise<PouchDB.Core.AllDocsResponse<Content & Model>>;
    /**
     * Create, update or delete multiple documents. The docs argument is an array of documents.
     * If you omit an _id parameter on a given document, the database will create a new document and assign the ID for you.
     * To update a document, you must include both an _id parameter and a _rev parameter,
     * which should match the ID and revision of the document on which to base your updates.
     * Finally, to delete a document, include a _deleted parameter with the value true.
     */
    bulkDocs<Model>(docs: Array<PouchDB.Core.PutDocument<Content & Model>>, options?: PouchDB.Core.BulkDocsOptions): Promise<SavedPouchDoc[]>;
    /** Compact the database */
    compact(options?: PouchDB.Core.CompactOptions): Promise<PouchDB.Core.Response>;
    /** Fetch a document */
    get<Model>(docId: PouchDB.Core.DocumentId, options?: PouchDB.Core.GetOptions): Promise<Content & Model & PouchDB.Core.IdMeta & PouchDB.Core.GetMeta>;
    /**
     * Create a new document without providing an id.
     *
     * You should prefer put() to post(), because when you post(), you are
     * missing an opportunity to use allDocs() to sort documents by _id
     * (because your _ids are random).
     *
     * @see {@link https://pouchdb.com/2014/06/17/12-pro-tips-for-better-code-with-pouchdb.html|PouchDB Pro Tips}
     */
    post<Model>(doc: PouchDB.Core.PostDocument<Content & Model>, options?: PouchDB.Core.Options): Promise<PouchDB.Core.Response>;
    /**
     * Create a new document or update an existing document.
     *
     * If the document already exists, you must specify its revision _rev,
     * otherwise a conflict will occur.
     * There are some restrictions on valid property names of the documents.
     * If you try to store non-JSON data (for instance Date objects) you may
     * see inconsistent results.
     */
    put<Model>(doc: PouchDB.Core.PutDocument<Content & Model>, options?: PouchDB.Core.PutOptions): Promise<PouchDB.Core.Response>;
    /** Remove a doc from the database */
    remove(doc: PouchDB.Core.RemoveDocument, options?: PouchDB.Core.Options): Promise<PouchDB.Core.Response>;
    /** Get database information */
    info(): Promise<PouchDB.Core.DatabaseInfo>;
    /**
     * A list of changes made to documents in the database, in the order they were made.
     * It returns an object with the method cancel(), which you call if you don’t want to listen to new changes anymore.
     *
     * It is an event emitter and will emit a 'change' event on each document change,
     * a 'complete' event when all the changes have been processed, and an 'error' event when an error occurs.
     * Calling cancel() will unsubscribe all event listeners automatically.
     */
    changes<Model>(options?: PouchDB.Core.ChangesOptions): PouchDB.Core.Changes<Content & Model>;
    /** Close the database */
    close(): Promise<void>;
    /**
     * Attaches a binary object to a document.
     * This method will update an existing document to add the attachment, so it requires a rev if the document already exists.
     * If the document doesn’t already exist, then this method will create an empty document containing the attachment.
     */
    putAttachment(docId: PouchDB.Core.DocumentId, attachmentId: PouchDB.Core.AttachmentId, attachment: PouchDB.Core.AttachmentData, type: string): Promise<PouchDB.Core.Response>;
    /** Get attachment data */
    getAttachment(docId: PouchDB.Core.DocumentId, attachmentId: PouchDB.Core.AttachmentId, options?: {
        rev?: PouchDB.Core.RevisionId;
    }): Promise<Blob | Buffer>;
    /** Delete an attachment from a doc. You must supply the rev of the existing doc. */
    removeAttachment(docId: PouchDB.Core.DocumentId, attachmentId: PouchDB.Core.AttachmentId, rev: PouchDB.Core.RevisionId): Promise<PouchDB.Core.RemoveAttachmentResponse>;
    /** Given a set of document/revision IDs, returns the document bodies (and, optionally, attachment data) for each ID/revision pair specified. */
    bulkGet<Model>(options: PouchDB.Core.BulkGetOptions): Promise<PouchDB.Core.BulkGetResponse<Content & Model>>;
    /** Given a set of document/revision IDs, returns the subset of those that do not correspond to revisions stored in the database */
    revsDiff(diff: PouchDB.Core.RevisionDiffOptions): Promise<PouchDB.Core.RevisionDiffResponse>;
}
export declare const pouchProxyMethods: string[];
export default Pouchy;
