"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint no-unused-vars: "off" */
/**
 * @module Pouchy
 */
const to_underscore_prefix_1 = require("./to-underscore-prefix");
const pouchdb_find_1 = __importDefault(require("pouchdb-find"));
const pouchdb_adapter_http_1 = __importDefault(require("pouchdb-adapter-http"));
const pouchdb_replication_1 = __importDefault(require("pouchdb-replication"));
const defaults_1 = __importDefault(require("lodash/defaults"));
const get_1 = __importDefault(require("lodash/get"));
const isNil_1 = __importDefault(require("lodash/isNil"));
const path_1 = __importDefault(require("path"));
const pouchdb_core_1 = __importDefault(require("pouchdb-core"));
const uniq_1 = __importDefault(require("lodash/uniq"));
const url_1 = __importDefault(require("url"));
pouchdb_core_1.default.plugin(pouchdb_adapter_http_1.default)
    .plugin(pouchdb_find_1.default)
    .plugin(pouchdb_replication_1.default);
exports.couchUrlify = (url) => url.replace(/[^/a-z0-9_$()+-]/gi, '');
exports.POUCHY_API_DOCS_URI = 'https://cdaringe.github.io/pouchy';
class Pouchy {
    constructor(opts) {
        // tap into your instance's replication `changes()` so you may listen to events.
        // calling `.destroy` will scrap this emitter. emitter only present when
        // `replicate` options intially provided
        this.syncEmitter = null;
        this._replicationOpts = null;
        this._validatePouchyOpts(opts);
        this.isEnforcingCouchDbSafe = isNil_1.default(opts.couchdbSafe)
            ? true
            : opts.couchdbSafe;
        this.url = this._getUrlFromOpts(opts);
        this.hasLocalDb = !!opts.name;
        /* istanbul ignore next */
        if (!this.url && !this.hasLocalDb) {
            throw new Error('remote database requires url');
        }
        this.name = this.hasLocalDb
            ? this._setDbNameFromOpts(opts)
            : this._setDbNameFromUri(this.url);
        if (this.isEnforcingCouchDbSafe)
            this._validateDbName();
        this.path = this.hasLocalDb
            ? path_1.default.resolve(opts.path || '', this.name)
            : null;
        this.db = new pouchdb_core_1.default(opts.name ? this.path : this.url, opts.pouchConfig);
        if (opts.replicate)
            this._handleReplication(opts.replicate);
    }
    /**
     * @private
     */
    _validatePouchyOpts(opts) {
        if (!opts || (!opts.name && (!opts.url && !opts.conn))) {
            throw new ReferenceError([
                'missing pouchy database paramters.  please see: ' +
                    exports.POUCHY_API_DOCS_URI +
                    '\n',
                '\tif you are creating a local database (browser or node), provide a `name` key.\n',
                '\tif you are just using pouchy to access a remote database, provide a `url` or `conn` key\n',
                '\tif you are creating a database to replicate with a remote database, provide a',
                '`url` or `conn` key, plus a replicate key.\n'
            ].join(''));
        }
        /* istanbul ignore next */
        if (opts.url && opts.conn) {
            throw new ReferenceError('provide only a `url` or `conn` option');
        }
        /* istanbul ignore next */
        if (!this) {
            throw new ReferenceError('no `this` context.  did you forget `new`?');
        }
    }
    /**
     * @private
     */
    _handleReplication(opts) {
        var replOpts;
        var mode;
        /* istanbul ignore next */
        if (!this.url) {
            throw new ReferenceError('url or conn object required to replicate');
        }
        /* istanbul ignore else */
        if (typeof opts === 'string') {
            mode = opts;
            replOpts = { live: true, retry: true };
        }
        else if (opts) {
            mode = Object.keys(opts)[0];
            /* istanbul ignore else */
            if (mode in opts) {
                replOpts = opts[mode];
            }
            else {
                throw new Error(`mode "${mode}" is not a valid replication option`);
            }
        }
        else {
            throw new Error('invalid replication options');
        }
        this._replicationOpts = replOpts;
        switch (mode) {
            /* istanbul ignore next */
            case 'out':
                this.syncEmitter = this.db.replicate.to(this.url, replOpts);
                break;
            /* istanbul ignore next */
            case 'in':
                this.syncEmitter = this.db.replicate.from(this.url, replOpts);
                break;
            case 'sync':
                this.syncEmitter = this.db.sync(this.url, replOpts);
                break;
            default:
                /* istanbul ignore next */
                throw new Error([
                    "in/out replication direction must be specified, got '",
                    mode + "'"
                ].join(' '));
        }
    }
    /**
     * @private
     */
    _getUrlFromOpts(opts) {
        if (!isNil_1.default(opts.url))
            return opts.url;
        if (!isNil_1.default(opts.conn))
            return url_1.default.format(opts.conn);
        return null;
    }
    /**
     * @private
     */
    _setDbNameFromUri(uri) {
        const pathname = url_1.default.parse(uri).pathname; // eslint-disable-line
        /* istanbul ignore next */
        if (!pathname && !this.name) {
            throw new Error([
                'unable to infer database name from uri. try adding a pathname',
                'to the uri (e.g. host.org/my-db-name) or pass a `name` option'
            ].join(' '));
        }
        var pathParts = (pathname || '').split('/');
        this.name = this.name || pathParts[pathParts.length - 1];
        return this.name;
    }
    /**
     * @private
     */
    _setDbNameFromOpts(opts) {
        /* istanbul ignore next */
        if (!opts.name) {
            throw new Error('local pouchy database requires a `name` field');
        }
        return opts.name;
    }
    /**
     * @private
     */
    _validateDbName() {
        var couchDbSafeName = exports.couchUrlify(this.name.toLowerCase());
        if (this.name === couchDbSafeName)
            return;
        throw new Error([
            'database name may not be couchdb safe.',
            '\tunsafe name: ' + this.name,
            '\tsafe name: ' + couchDbSafeName
        ].join('\n'));
    }
    /**
     * add a document to the db.
     * @see save
     * @example
     * // with _id
     * const doc = await p.add({ _id: 'my-sauce', bbq: 'sauce' })
     * console.log(doc._id, doc._rev, doc.bbq); // 'my-sauce', '1-a76...46c', 'sauce'
     *
     * // no _id
     * const doc = await p.add({ peanut: 'butter' })
     * console.log(doc._id, doc._rev, doc.peanut); // '66188...00BF885E', '1-0d74...7ac', 'butter'
     */
    add(doc) {
        return this.save.apply(this, arguments);
    }
    /**
     * get all documents from db
     * @example
     * const docs = await p.all()
     * console.log(`total # of docs: ${docs.length}!`))
     *
     * @example
     * const docs = await p.all({ includeDesignDocs: true })
     */
    all(allOpts) {
        return __awaiter(this, void 0, void 0, function* () {
            const opts = defaults_1.default(allOpts || {}, { include_docs: true });
            const docs = yield this.db.allDocs(opts);
            return docs.rows.reduce(function simplifyAllDocSet(r, v) {
                var doc = opts.include_docs ? v.doc : v;
                // rework doc format to always have id ==> _id
                if (!opts.include_docs) {
                    doc._id = doc.id;
                    doc._rev = doc.value.rev;
                    delete doc.id;
                    delete doc.value;
                    delete doc.key;
                }
                ;
                r.push(doc);
                return r;
            }, []);
        });
    }
    /**
     * The native bulkGet PouchDB API is not very user friendly.
     * In fact, it's down right wacky!
     * This method patches PouchDB's `bulkGet` and assumes that _all_ of your
     * requested docs exist.  If they do not, it will error via the usual error
     * control flows.
     *
     * @example
     * // A good example of what you can expect is actually right out of the tests!
     * let dummyDocs = [
     *   { _id: 'a', data: 'a' },
     *   { _id: 'b', data: 'b' }
     * ]
     * Promise.resolve()
     * .then(() => p.save(dummyDocs[0])) // add our first doc to the db
     * .then((doc) => (dummyDocs[0] = doc)) // update our closure doc it knows the _rev
     * .then(() => p.save(dummyDocs[1]))
     * .then((doc) => (dummyDocs[1] = doc))
     * .then(() => {
     *   // prepare getMany query (set of { _id, _rev}'s are required)
     *   const toFetch = dummyDocs.map(dummy => ({
     *     _id: dummy._id,
     *     _rev: dummy._rev
     *     // or you can provide .id, .rev
     *   }))
     *   p.getMany(toFetch)
     *     .then((docs) => {
     *       t.deepEqual(docs, dummyDocs, 'getMany returns sane results')
     *       t.end()
     *     })
     * })
     * @param {object|array} opts array of {_id, _rev}s, or { docs: [ ... } } where
     *                            ... is an array of {_id, _rev}s
     * @param {function} [cb]
     */
    getMany(docMetas) {
        return __awaiter(this, void 0, void 0, function* () {
            /* istanbul ignore else */
            if (!docMetas || !Array.isArray(docMetas)) {
                throw new Error('getMany: missing doc metadatas');
            }
            const opts = {
                docs: docMetas.map(function remapIdRev(lastDoc) {
                    const doc = Object.assign({}, lastDoc);
                    // we need to map back to id and rev here
                    /* istanbul ignore else */
                    if (doc._id)
                        doc.id = doc._id;
                    /* istanbul ignore else */
                    if (doc._rev)
                        doc.rev = doc._rev;
                    delete doc._rev;
                    delete doc._id;
                    return doc;
                })
            };
            if (!opts.docs.length)
                return Promise.resolve([]);
            const r = yield this.db.bulkGet(opts);
            return r.results.map(function tidyBulkGetDocs(docGroup) {
                var doc = get_1.default(docGroup, 'docs[0].ok');
                if (!doc) {
                    throw new ReferenceError('doc ' + docGroup.id + 'not found');
                }
                return doc;
            });
        });
    }
    /**
     * easy way to create a db index.
     * @see createIndicies
     * @example
     * await p.upsertIndex('myIndex')
     */
    upsertIndex(indexName) {
        return this.createIndicies.apply(this, arguments);
    }
    /**
     * allow single or bulk creation of indicies. also, doesn't flip out if you've
     * already set an index.
     * @example
     * const indicies = await p.createIndicies('test')
     * console.dir(indicies)
     * // ==>
     * [{
     *     id: "_design/idx-28933dfe7bc072c94e2646126133dc0d"
     *     name: "idx-28933dfe7bc072c94e2646126133dc0d"
     *     result: "created"
     * }]
     */
    createIndicies(indicies) {
        indicies = Array.isArray(indicies) ? indicies : [indicies];
        return (this.db
            .createIndex({ index: { fields: uniq_1.default(indicies) } })
            /* istanbul ignore next */
            .catch(function handleFailCreateIndicies(err) {
            /* istanbul ignore next */
            if (err.status !== 409)
                throw err;
        }));
    }
    /**
     * @see deleteAll
     * @returns {Promise}
     */
    clear() {
        return this.deleteAll.apply(this, arguments);
    }
    /**
     * delete a document.
     * @example
     * // same as pouch.remove
     * const deleted = await p.delete(doc)
     * console.dir(deleted)
     * // ==>
     * {
     *     id: "test-doc-1"
     *     ok: true
     *     rev: "2-5cf6a4725ed4b9398d609fc8d7af2553"
     * }
     */
    delete(doc, opts) {
        return this.db.remove(doc, opts);
    }
    /**
     * clears the db of documents. under the hood, `_deleted` flags are added to docs
     */
    deleteAll() {
        return __awaiter(this, void 0, void 0, function* () {
            var deleteSingleDoc = (doc) => this.delete(doc);
            const docs = yield this.all();
            return Promise.all(docs.map(deleteSingleDoc));
        });
    }
    /**
     * Destroys the database.  Proxies to pouchdb.destroy after completing
     * some internal cleanup first
     */
    destroy() {
        return __awaiter(this, void 0, void 0, function* () {
            /* istanbul ignore next */
            if (this.syncEmitter && !this.syncEmitter.canceled) {
                let isSyncCancelledP = Promise.resolve();
                if (this._replicationOpts && this._replicationOpts.live) {
                    // early bind the `complete` event listener.  careful not to bind it
                    // inside the .then, otherwise binding happens at the end of the event
                    // loop, which is too late! `.cancel` is a sync call!
                    isSyncCancelledP = new Promise((resolve, reject) => {
                        if (!this.syncEmitter) {
                            return reject(new Error('syncEmitter not found'));
                        }
                        this.syncEmitter.on('complete', () => {
                            resolve();
                        });
                    });
                }
                this.syncEmitter.cancel(); // will trigger a `complete` event
                yield isSyncCancelledP;
            }
            return this.db.destroy();
        });
    }
    /**
     * Similar to standard pouchdb.find, but returns simple set of results
     */
    findMany(opts) {
        return __awaiter(this, void 0, void 0, function* () {
            const rslt = yield this.db.find(opts);
            return rslt.docs;
        });
    }
    /**
     * update a document, and get your sensibly updated doc in return.
     *
     * @example
     * const doc = await p.update({ _id: 'my-doc', _rev: '1-abc123' })
     * console.log(doc)
     * // ==>
     * {
     *    _id: 'my-doc',
     *    _rev: '2-abc234'
     * }
     * @param {object} doc
     * @param {function} [cb]
     * @returns {Promise}
     */
    update(doc) {
        return __awaiter(this, void 0, void 0, function* () {
            // http://pouchdb.com/api.html#create_document
            // db.put(doc, [docId], [docRev], [options], [callback])
            const meta = yield this.db.put(doc);
            doc._id = meta.id;
            doc._rev = meta.rev;
            return doc;
        });
    }
    /**
     * Adds or updates a document.  If `_id` is set, a `put` is performed (basic add operation). If no `_id` present, a `post` is performed, in which the doc is added, and large-random-string is assigned as `_id`.
     * @example
     * const doc = await p.save({ beep: 'bop' })
     * console.log(doc)
     * // ==>
     * {
     *   _id: 'AFEALJW-234LKJASDF-2A;LKFJDA',
     *   _rev: '1-asdblkue242kjsa0f',
     *   beep: 'bop'
     * }
     */
    save(doc) {
        return __awaiter(this, void 0, void 0, function* () {
            // http://pouchdb.com/api.html#create_document
            // db.post(doc, [docId], [docRev], [options], [callback])
            /* istanbul ignore next */
            var method = Object.prototype.hasOwnProperty.call(doc, '_id') &&
                (doc._id || doc._id === 0)
                ? 'put'
                : 'post';
            const meta = method === 'put'
                ? yield this.db.put(doc)
                : yield this.db.post(doc);
            delete meta.status;
            doc._id = meta.id;
            doc._rev = meta.rev;
            return doc;
        });
    }
    /* istanbul ignore next */
    /** Fetch all documents matching the given options. */
    allDocs(options) {
        return {};
    }
    /* istanbul ignore next */
    /**
     * Create, update or delete multiple documents. The docs argument is an array of documents.
     * If you omit an _id parameter on a given document, the database will create a new document and assign the ID for you.
     * To update a document, you must include both an _id parameter and a _rev parameter,
     * which should match the ID and revision of the document on which to base your updates.
     * Finally, to delete a document, include a _deleted parameter with the value true.
     */
    bulkDocs(docs, options) {
        return {};
    }
    /* istanbul ignore next */
    /** Compact the database */
    compact(options) {
        return {};
    }
    /* istanbul ignore next */
    /** Fetch a document */
    get(docId, options) {
        return {};
    }
    /* istanbul ignore next */
    /**
     * Create a new document without providing an id.
     *
     * You should prefer put() to post(), because when you post(), you are
     * missing an opportunity to use allDocs() to sort documents by _id
     * (because your _ids are random).
     *
     * @see {@link https://pouchdb.com/2014/06/17/12-pro-tips-for-better-code-with-pouchdb.html|PouchDB Pro Tips}
     */
    post(doc, options) {
        return {};
    }
    /* istanbul ignore next */
    /**
     * Create a new document or update an existing document.
     *
     * If the document already exists, you must specify its revision _rev,
     * otherwise a conflict will occur.
     * There are some restrictions on valid property names of the documents.
     * If you try to store non-JSON data (for instance Date objects) you may
     * see inconsistent results.
     */
    put(doc, options) {
        return {};
    }
    /* istanbul ignore next */
    /** Remove a doc from the database */
    remove(doc, options) {
        return {};
    }
    /* istanbul ignore next */
    /** Get database information */
    info() {
        return {};
    }
    /* istanbul ignore next */
    /**
     * A list of changes made to documents in the database, in the order they were made.
     * It returns an object with the method cancel(), which you call if you don’t want to listen to new changes anymore.
     *
     * It is an event emitter and will emit a 'change' event on each document change,
     * a 'complete' event when all the changes have been processed, and an 'error' event when an error occurs.
     * Calling cancel() will unsubscribe all event listeners automatically.
     */
    changes(options) {
        return {};
    }
    /* istanbul ignore next */
    /** Close the database */
    close() {
        return {};
    }
    /* istanbul ignore next */
    /**
     * Attaches a binary object to a document.
     * This method will update an existing document to add the attachment, so it requires a rev if the document already exists.
     * If the document doesn’t already exist, then this method will create an empty document containing the attachment.
     */
    putAttachment(docId, attachmentId, attachment, type) {
        return {};
    }
    /* istanbul ignore next */
    /** Get attachment data */
    getAttachment(docId, attachmentId, options) {
        return {};
    }
    /* istanbul ignore next */
    /** Delete an attachment from a doc. You must supply the rev of the existing doc. */
    removeAttachment(docId, attachmentId, rev) {
        return {};
    }
    /* istanbul ignore next */
    /** Given a set of document/revision IDs, returns the document bodies (and, optionally, attachment data) for each ID/revision pair specified. */
    bulkGet(options) {
        return {};
    }
    /* istanbul ignore next */
    /** Given a set of document/revision IDs, returns the subset of those that do not correspond to revisions stored in the database */
    revsDiff(diff) {
        return {};
    }
}
exports.Pouchy = Pouchy;
Pouchy.PouchDB = pouchdb_core_1.default;
Pouchy.plugin = pouchdb_core_1.default.plugin;
Pouchy.defaults = pouchdb_core_1.default.defaults;
Pouchy.debug = pouchdb_core_1.default.debug;
/**
 * @private
 * call Pouchy or native PouchDB methods and transform all repsonses to
 * unify ids & revs
 */
/* istanbul ignore next */
const nonTransformingPrototype = {};
/* istanbul ignore next */
for (const key of Object.getOwnPropertyNames(Pouchy.prototype)) {
    const value = Pouchy.prototype[key];
    if (typeof value !== 'function' ||
        (value && value.name === 'Pouchy') ||
        (value && value.name && value.name[0] === '_')) {
        continue;
    }
    nonTransformingPrototype[key] = value;
    Pouchy.prototype[key] = function transformResponse(...args) {
        return __awaiter(this, void 0, void 0, function* () {
            let res = nonTransformingPrototype[key].call(this, ...args);
            if (res && res.then && res.catch)
                res = yield res;
            if (typeof args[args.length - 1] === 'function') {
                throw new Error([
                    'the pouchy-pouchdb callback interface has been removed.',
                    'please use the promise interface.'
                ].join(' '));
            }
            return to_underscore_prefix_1.toUnderscorePrefix(res);
        });
    };
}
exports.pouchProxyMethods = [
    'bulkDocs',
    'bulkGet',
    'changes',
    'close',
    'compact',
    'get',
    'getAttachment',
    'info',
    'post',
    'put',
    'putAttachment',
    'remove',
    'removeAttachment',
    'revsDiff'
];
for (const key of exports.pouchProxyMethods) {
    const value = pouchdb_core_1.default.prototype[key];
    /* istanbul ignore next */
    if (!value)
        throw new Error(`pouchdb method "${key}" not found`);
    /* istanbul ignore next */
    if (typeof value !== 'function')
        continue;
    Pouchy.prototype[key] = function proxyAndTransform(...args) {
        return __awaiter(this, void 0, void 0, function* () {
            const pouchMethod = pouchdb_core_1.default.prototype[key];
            let res = pouchMethod.call(this.db, ...args);
            if (res && res.then && res.catch)
                res = yield res;
            /* istanbul ignore next */
            if (typeof args[args.length - 1] === 'function') {
                throw new Error([
                    'the pouchy-pouchdb callback interface has been removed.',
                    'please use the promise interface.'
                ].join(' '));
            }
            return to_underscore_prefix_1.toUnderscorePrefix(res);
        });
    };
}
exports.default = Pouchy;
//# sourceMappingURL=index.js.map